// ----------------------------------
//        The N-Queens Problem
// ----------------------------------
//   10 queens -->    724 answers
//   11 queens -->   2680 answers
//   12 queens -->  14200 answers
//   13 queens -->  73712 answers
//   14 queens --> 365596 answers
// ----------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include "spinlock.h"

#define ABS(N) ((N) > 0 ? (N) : -(N))
#define MMAP_QUEUE_SIZE (queens == 10 ?  163840 : queens == 11 ?   819200 : \
                         queens == 12 ? 4915200 : queens == 13 ? 34406400 : 275251200)

int workers, workers_per_group, groups, queens, show, answers_per_worker;

typedef char board;

typedef struct {
  spinlock_t lock_answers;         // spinlock to access 'answers'
  int answers;                     // total number of answers found
  spinlock_t lock_idle_groups;     // spinlock to access 'idle_groups'
  int idle_groups;                 // number of groups without work
  spinlock_t lock_finish_workers;  // spinlock to access 'finish_workers'
  int finish_workers;              // number of workers that have already finished
} tasks_globals;
tasks_globals *my_globals;

#define lock_answers        (my_globals->lock_answers)
#define answers             (my_globals->answers)
#define lock_idle_groups    (my_globals->lock_idle_groups)
#define idle_groups         (my_globals->idle_groups)
#define lock_finish_workers (my_globals->lock_finish_workers)
#define finish_workers      (my_globals->finish_workers)

typedef struct {
  spinlock_t lock_queue; // spinlock to access queue of tasks
  int idle_workers;      // number of idle workers in group
  int next_put_group;    // next group where to put tasks
  board *queue_start;    // pointer to beginning of queue of tasks
  board *queue_end;      // pointer to end of queue of tasks
  board *get_task;       // pointer to older task
  board *put_task;       // pointer to younger task
} tasks_group;
tasks_group *my_groups;

#define lock_queue(GROUP)      (my_groups[GROUP].lock_queue)
#define idle_workers(GROUP)    (my_groups[GROUP].idle_workers)
#define next_put_group(GROUP)  (my_groups[GROUP].next_put_group)
#define queue_start(GROUP)     (my_groups[GROUP].queue_start)
#define queue_end(GROUP)       (my_groups[GROUP].queue_end)
#define get_task(GROUP)        (my_groups[GROUP].get_task)
#define put_task(GROUP)        (my_groups[GROUP].put_task)

void do_work(int);
int get_work(int, board *);
void put_work(int, board *);
void show_board(board *);
double current_time(void);
void *mmap_init(char *, int);
void parse_argv(int, char **);


int main(int argc, char *argv[]) {
  double execution_time;
  int i;

  // default values
  workers_per_group = 1;
  groups = 1;
  queens = 10;
  show = 0;
  // command line parsing
  parse_argv(argc, argv);
  workers = groups * workers_per_group;
  answers_per_worker = 0;
  // map file 'tasks_queue' into a shared memory region
  my_globals = (tasks_globals *) mmap_init("tasks_queue", sizeof(tasks_globals) + groups * sizeof(tasks_group) + MMAP_QUEUE_SIZE);
  // init shared variables
  spin_lock_init(lock_answers);
  answers = 0;
  spin_lock_init(lock_idle_groups);
  idle_groups = 0;
  spin_lock_init(lock_finish_workers);
  finish_workers = 0;
  // init group data structures
  my_groups = (tasks_group *)(my_globals + 1);
  queue_start(0) = (board *)(my_groups + groups);
  for (i = 0; i < groups; i++) {
    spin_lock_init(lock_queue(i));
    idle_workers(i) = 0;
    next_put_group(i) = (i + 1) % groups;
    queue_start(i) = queue_start(0) + i * MMAP_QUEUE_SIZE / groups;
    queue_end(i) = queue_start(i) + MMAP_QUEUE_SIZE / groups - queens;
    get_task(i) = put_task(i) = queue_start(i);
  }
  // initial task
  for(i = 0; i < queens; i++)
    get_task(0)[i] = -1;
  put_task(0) += queens;
  // create workers and execute
  execution_time = current_time();
  for (i = 1; i < workers; i++)
    if (fork() == 0) {
      do_work(i);
      return 0;
    }
  do_work(0);
  while (finish_workers != workers);
  execution_time = current_time() - execution_time;
  // show results
  printf("Total answers found: %d\n", answers);
  printf("Total execution time: %f seconds\n", execution_time);
  return 0;
}

void do_work(int worker) {
  int my_group, len, row, col, rowdist, coldist;
  board my_board[queens];
  my_group = worker / workers_per_group;
  while (1) {
    // get valid intermediate board
    if (!get_work(my_group, my_board)) {
      // no more work --> update total answers and finish execution
      spin_lock(lock_finish_workers);
      printf("Worker %d answers found: %d\n", worker, answers_per_worker);
      answers += answers_per_worker;
      finish_workers++;
      spin_unlock(lock_finish_workers);
      return;
    }
    len = 0;
    while (my_board[len] != -1)
      len++;
    // generate new intermediate boards
    for (col = 0; col < queens; col++) {
      my_board[len] = col;
      for (row = 0; row < len; row++) {
	rowdist = ABS(len - row);
	coldist = ABS(col - my_board[row]);
	if (coldist == 0 || coldist == rowdist)
	  break;
      }
      // check if valid intermediate board
      if (row == len) {
	if (len + 1 == queens){
	  answers_per_worker++;
	  if (show)
	    show_board(my_board);
	}
	else
	  put_work(my_group, my_board);
      }
    }
  }
  return;
}

int get_work(int my_group, board *my_board) {
  ...
}

void put_work(int my_group, board *my_board) {
  ...
}

void show_board(board *my_board) {
  int i;
  spin_lock(lock_answers);
  for (i = 0; i < queens; i++)
    if (my_board[i] > 9)
      printf("1%c ", '0' + my_board[i] - 10);
    else
      printf("0%c ", '0' + my_board[i]);
  printf("\n");
  spin_unlock(lock_answers);
  return;
}

double current_time(void) {
  struct timeval tempo;
  gettimeofday(&tempo, NULL);
  return ((double)tempo.tv_sec + (double)tempo.tv_usec / 1000000);
}

void *mmap_init(char *file, int size) {
  int fd;
  void *addr;
  if ((fd = open(file, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR)) == -1) {
    printf("open error!\n");
    exit(0);
  }
  if (lseek(fd, size, SEEK_SET) == -1) {
    printf("lseek error!\n");
    exit(0);
  }
  if (write(fd, "", 1) == -1) {
    printf("write error!\n");
    exit(0);
  }
  if ((addr = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0)) == (void *)-1) {
    printf("mmap error!\n");
    exit(0);
  }
  return addr;
}

void parse_argv(int argc, char *argv[]) {
  int i;
  for (i = 1; i < argc; i++) {
    if (argv[i][0] == '-') {
      if (argv[i][1] == 'q') {
        queens = atoi(&argv[i][2]);
      } else if (argv[i][1] == 'w') {
        workers_per_group = atoi(&argv[i][2]);
      } else if (argv[i][1] == 'g') {
        groups = atoi(&argv[i][2]);
      } else if (argv[i][1] == 's') {
	show = 1;
      } else {
        printf("nqueens: invalid argument (%s)\n", argv[i]);
	printf("Usage: nqueens [-gGROUPS] [-wWORKERS_PER_GROUP] [-qQUEENS] [-s]\n");
        exit(1);
      }
    } else {
      printf("nqueens: invalid argument (%s)\n", argv[i]);
      printf("Usage: nqueens [-gGROUPS] [-wWORKERS_PER_GROUP] [-qQUEENS] [-s]\n");
      exit(1);
    }
  }
  printf("The %2d-queens problem: %d group(s) with %d worker(s) each\n", queens, groups, workers_per_group);
  return;
}
